import * as util from "util";
import * as fs from "fs";
import { EventEmitter } from "events";
import * as path from "path";
import { Lock, Observable } from "@hibas123/utils";
import { Adapter, LoggingTypes, Message } from "./types";
import { ConsoleWriter } from "./consolewriter";


export interface LoggingBaseOptions {
    /**
     * Name will be prefixed on Console output and added to logfiles, if not specified here
     */
    name: string,
    files: boolean | {
        /**
         * Filename/path of the logfile. Skip if generated with name
         */
        logfile: string;
        /**
         * Filename/path of the logfile. Skip if generated with name
         */
        errorfile: string;
    }
    /**
     * Prints output to console
     */
    console: boolean;
}

export class LoggingBase {
    private logFile: any;
    private errorFile: any;
    private adapter: Adapter[] = [];
    private adapter_init: Promise<void>[] = [];

    private messageObservable = new Observable<Message>();
    private name: string;

    constructor(options?: Partial<LoggingBaseOptions> | string) {
        let opt: Partial<LoggingBaseOptions>;
        if (!options) opt = {}
        else if (typeof options === "string") {
            opt = { name: options };
        } else {
            opt = options;
        }

        let config = {
            name: undefined,
            console: true,
            files: true,
            ...opt
        };

        if (typeof config.files === "boolean") {
            if (config.files === true) {
                let name = "";
                if (config.name)
                    name = "." + config.name;
                config.files = {
                    logfile: `./logs/all${name}.log`,
                    errorfile: `./logs/error${name}.log`
                }
            } else {
                config.files = undefined;
            }
        }

        if (config.name)
            this.name = config.name;

        for (let key in this) {
            if (typeof this[key] === "function") this[key] = (<any>this[key]).bind(this);
        }

        if (typeof config.files !== "boolean" && config.files !== undefined && (config.files.logfile || config.files.errorfile)) {
            const { LoggingFiles } = require("./filewriter");
            if (config.files.logfile) {
                this.addAdapter(new LoggingFiles(config.files.logfile));
            }

            if (config.files.errorfile) {
                this.addAdapter(new LoggingFiles(config.files.errorfile, true));
            }
        }

        if (config.console) {
            this.addAdapter(new ConsoleWriter());
        }
    }

    addAdapter(adapter: Adapter) {
        this.adapter.push(adapter);
        let prms = Promise.resolve(adapter.init(this.messageObservable.getPublicApi(), this.name));
        this.adapter_init.push(prms);
    }

    flush(sync: true): void;
    flush(sync: false): Promise<void>;
    flush(sync: boolean): void | Promise<void> {
        if (sync) {
            this.adapter.forEach(elm => elm.flush(true));
        } else {
            return Promise.all(this.adapter.map(elm => elm.flush(false))).then(() => { });
        }
    }

    public waitForSetup() {
        return Promise.all(this.adapter_init);
    }

    public events: EventEmitter = new EventEmitter();

    debug(...message: any[]) {
        this.message(LoggingTypes.Debug, message);
    }

    log(...message: any[]) {
        this.message(LoggingTypes.Log, message);
    }

    warning(...message: any[]) {
        this.message(LoggingTypes.Warning, message);
    }

    logWithCustomColors(type: LoggingTypes, colors: string, ...message: any[]) {
        this.message(type, message, colors);
    }

    error(error: Error | string) {
        if (!error) error = "Empty ERROR was passed, so no informations available";
        if (typeof error === "string") {
            let e = new Error()
            this.message(LoggingTypes.Error, [error, "\n", e.stack]);
        } else {
            this.message(LoggingTypes.Error, [error.message, "\n", error.stack], undefined, getCallerFromExisting(error));
        }
    }

    errorMessage(...message: any[]) {
        this.message(LoggingTypes.Error, message);
    }

    private message(type: LoggingTypes, message: any[] | string, customColors?: string, caller?: { file: string, line: number }) {
        let file_raw = caller || getCallerFile();
        let file = `${file_raw.file}:${String(file_raw.line).padEnd(3, " ")}`;

        let mb = "";
        if (typeof message === "string") {
            mb = message;
        } else {
            message.forEach((e, i) => {
                if (typeof e !== "string") e = util.inspect(e, false, null);
                if (e.endsWith("\n") || i === message.length - 1) {
                    mb += e;
                } else {
                    mb += e + " ";
                }
            });
        }

        let lines = mb.split("\n");

        let date = new Date().toISOString().replace(/T/, ' ').replace(/\..+/, '');
        let prefix = `[ ${date} ][${LoggingTypes[type].toUpperCase().padEnd(5, " ")}][${file}]: `;

        let formatted = lines.map(line => prefix + line);

        let msg: Message = {
            date: new Date(),
            file,
            name: this.name,
            text: {
                raw: lines,
                formatted
            },
            type,
            customColors
        }

        this.messageObservable.send(msg);
    }
}