"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@hibas123/utils");
const consolewriter_1 = require("./consolewriter");
const inspect_1 = require("./inspect");
const types_1 = require("./types");
exports.Colors = {
    Reset: "\x1b[0m",
    Bright: "\x1b[1m",
    Dim: "\x1b[2m",
    Underscore: "\x1b[4m",
    Blink: "\x1b[5m",
    Reverse: "\x1b[7m",
    Hidden: "\x1b[8m",
    FgBlack: "\x1b[30m",
    FgRed: "\x1b[31m",
    FgGreen: "\x1b[32m",
    FgYellow: "\x1b[33m",
    FgBlue: "\x1b[34m",
    FgMagenta: "\x1b[35m",
    FgCyan: "\x1b[36m",
    FgWhite: "\x1b[37m",
    BgBlack: "\x1b[40m",
    BgRed: "\x1b[41m",
    BgGreen: "\x1b[42m",
    BgYellow: "\x1b[43m",
    BgBlue: "\x1b[44m",
    BgMagenta: "\x1b[45m",
    BgCyan: "\x1b[46m",
    BgWhite: "\x1b[47m"
};
class LoggingBase {
    constructor(options) {
        this.adapter = [];
        this.adapter_init = [];
        this.messageObservable = new utils_1.Observable();
        let opt;
        if (!options)
            opt = {};
        else if (typeof options === "string") {
            opt = { name: options };
        }
        else {
            opt = options;
        }
        let config = Object.assign({ name: undefined, console: true, files: true }, opt);
        if (config.name)
            this.name = config.name;
        for (let key in this) {
            if (typeof this[key] === "function")
                this[key] = this[key].bind(this);
        }
        if (config.console) {
            this.addAdapter(new consolewriter_1.ConsoleWriter());
        }
    }
    addAdapter(adapter) {
        this.adapter.push(adapter);
        let prms = Promise.resolve(adapter.init(this.messageObservable.getPublicApi(), this.name));
        this.adapter_init.push(prms);
    }
    flush(sync) {
        if (sync) {
            this.adapter.forEach(elm => elm.flush(true));
        }
        else {
            return Promise.all(this.adapter.map(elm => elm.flush(false))).then(() => { });
        }
    }
    waitForSetup() {
        return Promise.all(this.adapter_init);
    }
    debug(...message) {
        this.message(types_1.LoggingTypes.Debug, message);
    }
    log(...message) {
        this.message(types_1.LoggingTypes.Log, message);
    }
    warning(...message) {
        this.message(types_1.LoggingTypes.Warning, message);
    }
    logWithCustomColors(type, colors, ...message) {
        this.message(type, message, colors);
    }
    error(error) {
        if (!error)
            error = "Empty ERROR was passed, so no informations available";
        if (typeof error === "string") {
            let e = new Error();
            this.message(types_1.LoggingTypes.Error, [error, "\n", e.stack]);
        }
        else {
            this.message(types_1.LoggingTypes.Error, [error.message, "\n", error.stack], undefined, getCallerFromExisting(error));
        }
    }
    errorMessage(...message) {
        this.message(types_1.LoggingTypes.Error, message);
    }
    message(type, message, customColors, caller) {
        let file_raw = caller || getCallerFile();
        let file = `${file_raw.file}:${String(file_raw.line).padEnd(3, " ")}`;
        let mb = "";
        if (typeof message === "string") {
            mb = message;
        }
        else {
            message.forEach((e, i) => {
                if (typeof e !== "string")
                    e = inspect_1.default(e, { colors: true, showHidden: true, depth: 3 });
                if (e.endsWith("\n") || i === message.length - 1) {
                    mb += e;
                }
                else {
                    mb += e + " ";
                }
            });
        }
        let lines = mb.split("\n");
        let date = new Date().toISOString().replace(/T/, ' ').replace(/\..+/, '');
        let prefix = `[ ${date} ][${types_1.LoggingTypes[type].toUpperCase().padEnd(5, " ")}][${file}]: `;
        let formatted = lines.map(line => prefix + line);
        let msg = {
            date: new Date(),
            file,
            name: this.name,
            text: {
                raw: lines,
                formatted
            },
            type,
            customColors
        };
        this.messageObservable.send(msg);
    }
}
exports.LoggingBase = LoggingBase;
function getStack() {
    // Save original Error.prepareStackTrace
    let origPrepareStackTrace = Error.prepareStackTrace;
    // Override with function that just returns `stack`
    Error.prepareStackTrace = function (_, stack) {
        return stack;
    };
    // Create a new `Error`, which automatically gets `stack`
    let err = new Error();
    // Evaluate `err.stack`, which calls our new `Error.prepareStackTrace`
    let stack = err.stack;
    // Restore original `Error.prepareStackTrace`
    Error.prepareStackTrace = origPrepareStackTrace;
    // Remove superfluous function call on stack
    stack.shift(); // getStack --> Error
    return stack;
}
function baseName(path) {
    return path.split(/[\\/]/).pop();
}
function getCallerFile() {
    try {
        let stack = getStack();
        let current_file = stack.shift().getFileName();
        while (stack.length) {
            let caller_file = stack.shift();
            const util = require("util");
            if (current_file !== caller_file.getFileName())
                return {
                    file: baseName(caller_file.getFileName()),
                    line: caller_file.getLineNumber()
                };
        }
    }
    catch (err) { }
    return { file: undefined, line: 0 };
}
function getCallerFromExisting(err) {
    if (!err || !err.stack)
        return { file: "NOFILE", line: 0 };
    let lines = err.stack.split("\n");
    lines.shift(); // removing first line
    while (lines.length > 0) {
        let line = lines.shift();
        let matches = line.match(/[a-zA-Z_-]+[.][a-zA-Z_-]+[:][0-9]+/g);
        if (matches && matches.length > 0) {
            let [f, line] = matches[0].split(":");
            return {
                file: f, line: Number(line)
            };
        }
    }
}
//# sourceMappingURL=base.js.map